# Copyright 2016 Hewlett Packard Enterprise Development, LP
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import segment as extension
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants
from neutron_lib.db import api as db_api
from neutron_lib.db import resource_extend
from neutron_lib.db import utils as db_utils
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import directory
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_utils import uuidutils

from neutron.db.models import agent as agent_model
from neutron.db.models import segment as segment_model
from neutron.db import segments_db as db
from neutron import manager
from neutron.objects import base as base_obj
from neutron.objects import network
from neutron.services.segments import exceptions


LOG = logging.getLogger(__name__)
_USER_CONFIGURED_SEGMENT_PLUGIN = None
FOR_NET_DELETE = 'for_net_delete'


def check_user_configured_segment_plugin():
    global _USER_CONFIGURED_SEGMENT_PLUGIN
    # _USER_CONFIGURED_SEGMENT_PLUGIN will contain 3 possible values:
    # 1. None, this just happens during neutron-server startup.
    # 2. True, this means that users configure the 'segments'
    #    service plugin in neutron config file.
    # 3. False, this means that can not find 'segments' service
    #    plugin in neutron config file.
    # This function just load once to store the result
    # into _USER_CONFIGURED_SEGMENT_PLUGIN during neutron-server startup.
    if _USER_CONFIGURED_SEGMENT_PLUGIN is None:
        segment_class = 'neutron.services.segments.plugin.Plugin'
        _USER_CONFIGURED_SEGMENT_PLUGIN = any(
            p in cfg.CONF.service_plugins for p in ['segments', segment_class])
    return _USER_CONFIGURED_SEGMENT_PLUGIN


class SegmentDbMixin(object):
    """Mixin class to add segment."""

    @staticmethod
    def _make_segment_dict(segment_obj, fields=None):
        res = {'id': segment_obj['id'],
               'network_id': segment_obj['network_id'],
               'name': segment_obj['name'],
               'description': segment_obj['description'],
               db.PHYSICAL_NETWORK: segment_obj[db.PHYSICAL_NETWORK],
               db.NETWORK_TYPE: segment_obj[db.NETWORK_TYPE],
               db.SEGMENTATION_ID: segment_obj[db.SEGMENTATION_ID],
               'hosts': segment_obj['hosts'],
               'segment_index': segment_obj['segment_index']}
        resource_extend.apply_funcs('segments', res, segment_obj.db_obj)
        return db_utils.resource_fields(res, fields)

    def _get_segment(self, context, segment_id):
        segment = network.NetworkSegment.get_object(context, id=segment_id)
        if not segment:
            raise exceptions.SegmentNotFound(segment_id=segment_id)
        return segment

    @log_helpers.log_method_call
    def create_segment(self, context, segment):
        """Create a segment."""
        segment = segment['segment']
        segment_id = segment.get('id') or uuidutils.generate_uuid()
        try:
            new_segment = self._create_segment_db(context, segment_id, segment)
        except db_exc.DBReferenceError:
            raise n_exc.NetworkNotFound(net_id=segment['network_id'])
        registry.publish(resources.SEGMENT, events.AFTER_CREATE, self,
                         payload=events.DBEventPayload(
                             context, resource_id=segment_id,
                             states=(new_segment,)))
        return self._make_segment_dict(new_segment)

    def _create_segment_db(self, context, segment_id, segment):
        with db_api.CONTEXT_WRITER.using(context):
            network_id = segment['network_id']
            physical_network = segment[extension.PHYSICAL_NETWORK]
            if physical_network == constants.ATTR_NOT_SPECIFIED:
                physical_network = None
            network_type = segment[extension.NETWORK_TYPE]
            segmentation_id = segment[extension.SEGMENTATION_ID]
            if segmentation_id == constants.ATTR_NOT_SPECIFIED:
                segmentation_id = None
            name = segment['name']
            if name == constants.ATTR_NOT_SPECIFIED:
                name = None
            description = segment['description']
            if description == constants.ATTR_NOT_SPECIFIED:
                description = None
            args = {'id': segment_id,
                    'network_id': network_id,
                    'name': name,
                    'description': description,
                    db.PHYSICAL_NETWORK: physical_network,
                    db.NETWORK_TYPE: network_type,
                    db.SEGMENTATION_ID: segmentation_id}
            # Calculate the index of segment
            segment_index = 0
            segments = self.get_segments(
                context,
                filters={'network_id': [network_id]},
                fields=['segment_index'],
                sorts=[('segment_index', True)])
            if segments:
                # NOTE(xiaohhui): The new index is the last index + 1, this
                # may cause discontinuous segment_index. But segment_index
                # can functionally work as the order index for segments.
                segment_index = (segments[-1].get('segment_index') + 1)
            args['segment_index'] = segment_index

            new_segment = network.NetworkSegment(context, **args)
            new_segment.create()
            # Do some preliminary operations before committing the segment to
            # db
            registry.publish(
                resources.SEGMENT, events.PRECOMMIT_CREATE, self,
                payload=events.DBEventPayload(context, resource_id=segment_id,
                                              states=(new_segment,)))
            # The new segment might have been updated by the callbacks
            # subscribed to the PRECOMMIT_CREATE event. So update it in the DB
            new_segment.update()
            return new_segment

    @log_helpers.log_method_call
    def update_segment(self, context, uuid, segment):
        """Update an existing segment."""
        segment = segment['segment']
        with db_api.CONTEXT_WRITER.using(context):
            curr_segment = self._get_segment(context, uuid)
            curr_segment.update_fields(segment)
            curr_segment.update()
        return self._make_segment_dict(curr_segment)

    @log_helpers.log_method_call
    def get_segment(self, context, uuid, fields=None):
        segment_db = self._get_segment(context, uuid)
        return self._make_segment_dict(segment_db, fields)

    @log_helpers.log_method_call
    def get_segments(self, context, filters=None, fields=None,
                     sorts=None, limit=None, marker=None,
                     page_reverse=False):
        filters = filters or {}
        pager = base_obj.Pager(sorts, limit, page_reverse, marker)
        segment_objs = network.NetworkSegment.get_objects(
            context, _pager=pager, **filters)
        return [self._make_segment_dict(obj) for obj in segment_objs]

    @log_helpers.log_method_call
    def get_segments_count(self, context, filters=None):
        filters = filters or {}
        return network.NetworkSegment.count(context, **filters)

    @log_helpers.log_method_call
    def get_segments_by_hosts(self, context, hosts):
        if not hosts:
            return []
        segment_host_mapping = network.SegmentHostMapping.get_objects(
            context, host=hosts)
        return list({mapping.segment_id for mapping in segment_host_mapping})

    @log_helpers.log_method_call
    def delete_segment(self, context, uuid, for_net_delete=False):
        """Delete an existing segment."""
        segment_dict = self.get_segment(context, uuid)
        # Do some preliminary operations before deleting the segment
        registry.publish(resources.SEGMENT, events.BEFORE_DELETE,
                         self.delete_segment,
                         payload=events.DBEventPayload(
                             context, metadata={
                                 FOR_NET_DELETE: for_net_delete},
                             states=(segment_dict,),
                             resource_id=uuid))

        # Delete segment in DB
        with db_api.CONTEXT_WRITER.using(context):
            if not network.NetworkSegment.delete_objects(context, id=uuid):
                raise exceptions.SegmentNotFound(segment_id=uuid)
            # Do some preliminary operations before deleting segment in db
            registry.publish(resources.SEGMENT, events.PRECOMMIT_DELETE,
                             self.delete_segment,
                             payload=events.DBEventPayload(
                                 context, metadata={
                                     FOR_NET_DELETE: for_net_delete},
                                 resource_id=uuid,
                                 states=(segment_dict,)))

        registry.publish(resources.SEGMENT, events.AFTER_DELETE,
                         self.delete_segment,
                         payload=events.DBEventPayload(
                             context, metadata={
                                 FOR_NET_DELETE: for_net_delete},
                             states=(segment_dict,),
                             resource_id=uuid))


@db_api.retry_if_session_inactive()
@lockutils.synchronized('update_segment_host_mapping')
def update_segment_host_mapping(context, host, current_segment_ids):
    with db_api.CONTEXT_WRITER.using(context):
        segment_host_mapping = network.SegmentHostMapping.get_objects(
            context, host=host)
        previous_segment_ids = {
            seg_host['segment_id'] for seg_host in segment_host_mapping}
        segment_ids = current_segment_ids - previous_segment_ids
        for segment_id in segment_ids:
            network.SegmentHostMapping(
                context, segment_id=segment_id, host=host).create()
        LOG.debug('Segments %s mapped to the host %s', segment_ids, host)
        stale_segment_ids = previous_segment_ids - current_segment_ids
        if stale_segment_ids:
            for entry in segment_host_mapping:
                if entry.segment_id in stale_segment_ids:
                    entry.delete()
                    LOG.debug('Segment %s unmapped from host %s',
                              entry.segment_id, entry.host)


def get_hosts_mapped_with_segments(context, include_agent_types=None,
                                   exclude_agent_types=None):
    """Get hosts that are mapped with segments.

    L2 providers can use this method to get an overview of SegmentHostMapping,
    and then delete the stale SegmentHostMapping.

    When using both include_agent_types and exclude_agent_types,
    exclude_agent_types is most significant.
    All hosts without agent are excluded when using any agent_type filter.

    :param context: current running context information
    :param include_agent_types: (set) List of agent types, include hosts
        with matching agents.
    :param exclude_agent_types: (set) List of agent types, exclude hosts
        with matching agents.
    """
    def add_filter_by_agent_types(qry, include, exclude):
        qry = qry.join(
            agent_model.Agent,
            segment_model.SegmentHostMapping.host == agent_model.Agent.host)
        if include:
            qry = qry.filter(agent_model.Agent.agent_type.in_(include))
        if exclude:
            qry = qry.filter(agent_model.Agent.agent_type.not_in(exclude))

        return qry

    with db_api.CONTEXT_READER.using(context):
        query = context.session.query(segment_model.SegmentHostMapping)
        if include_agent_types or exclude_agent_types:
            query = add_filter_by_agent_types(query, include_agent_types,
                                              exclude_agent_types)

        res = query.all()

    return {row.host for row in res}


def _get_phys_nets(agent):
    configurations_dict = agent.get('configurations', {})
    mappings = configurations_dict.get('bridge_mappings', {})
    mappings.update(configurations_dict.get('interface_mappings', {}))
    mappings.update(configurations_dict.get('device_mappings', {}))
    return list(mappings.keys())


reported_hosts = set()

# NOTE: Module level variable of segments plugin. It should be removed once
# segments becomes a default plugin.
segments_plugin = None


def get_segments_with_phys_nets(context, phys_nets):
    """Get segments from physical networks.

    L2 providers usually have information of hostname and physical networks.
    They could use this method to get related segments and then update
    SegmentHostMapping.
    """
    phys_nets = list(phys_nets)
    if not phys_nets:
        return []

    with db_api.CONTEXT_READER.using(context):
        return network.NetworkSegment.get_objects(
            context, physical_network=phys_nets)


def map_segment_to_hosts(context, segment_id, hosts):
    """Map segment to a collection of hosts."""
    with db_api.CONTEXT_WRITER.using(context):
        for host in hosts:
            network.SegmentHostMapping(
                context, segment_id=segment_id, host=host).create()
    LOG.debug('Segment %s mapped to the hosts %s', segment_id, hosts)


def _update_segment_host_mapping_for_agent(resource, event, trigger,
                                           payload=None):
    plugin = payload.metadata.get('plugin')
    agent = payload.desired_state
    host = payload.metadata.get('host')
    context = payload.context

    check_segment_for_agent = getattr(plugin, 'check_segment_for_agent', None)
    if (not check_user_configured_segment_plugin() or
            not check_segment_for_agent):
        return
    phys_nets = _get_phys_nets(agent)
    if not phys_nets:
        return
    start_flag = agent.get('start_flag', None)
    if host in reported_hosts and not start_flag:
        return
    reported_hosts.add(host)
    if (len(payload.states) > 1 and
            payload.states[1] is not None and
            agent.get('configurations') == payload.states[1].get(
                'configurations')):
        return
    segments = get_segments_with_phys_nets(context, phys_nets)
    current_segment_ids = {
        segment['id'] for segment in segments
        if check_segment_for_agent(segment, agent)}
    update_segment_host_mapping(context, host, current_segment_ids)
    registry.publish(resources.SEGMENT_HOST_MAPPING, events.AFTER_CREATE,
                     plugin, payload=events.DBEventPayload(
                         context,
                         metadata={
                             'host': host,
                             'current_segment_ids': current_segment_ids}))


def _add_segment_host_mapping_for_segment(resource, event, trigger,
                                          payload=None):
    context = payload.context
    segment = payload.latest_state
    if not segment.physical_network:
        return
    cp = directory.get_plugin()
    check_segment_for_agent = getattr(cp, 'check_segment_for_agent', None)
    if not check_user_configured_segment_plugin() or not hasattr(
            cp, 'get_agents') or not check_segment_for_agent:
        # not an agent-supporting plugin
        registry.unsubscribe(_add_segment_host_mapping_for_segment,
                             resources.SEGMENT, events.PRECOMMIT_CREATE)
        return
    hosts = {agent['host'] for agent in cp.get_agents(context)
             if check_segment_for_agent(segment, agent)}
    map_segment_to_hosts(context, segment.id, hosts)


def _delete_segments_for_network(resource, event, trigger,
                                 payload=None, **kwargs):
    network_id = payload.resource_id
    admin_ctx = payload.context.elevated()
    global segments_plugin
    if not segments_plugin:
        segments_plugin = manager.NeutronManager.load_class_for_provider(
            'neutron.service_plugins', 'segments')()
    segments = segments_plugin.get_segments(
        admin_ctx, filters={'network_id': [network_id]})
    for segment in segments:
        segments_plugin.delete_segment(admin_ctx, segment['id'],
                                       for_net_delete=True)


def subscribe():
    registry.subscribe(_update_segment_host_mapping_for_agent,
                       resources.AGENT,
                       events.AFTER_CREATE)
    registry.subscribe(_update_segment_host_mapping_for_agent,
                       resources.AGENT,
                       events.AFTER_UPDATE)
    registry.subscribe(_add_segment_host_mapping_for_segment,
                       resources.SEGMENT, events.PRECOMMIT_CREATE)
    registry.subscribe(_delete_segments_for_network,
                       resources.NETWORK,
                       events.PRECOMMIT_DELETE)


subscribe()



# Copyright 2022 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

from neutron.agent.linux import utils as linux_utils
from neutron.privileged.agent.linux import devlink as priv_devlink


PortIndex = collections.namedtuple(
    'PortIndex', ['name', 'pf_pci', 'pf_num', 'vf_num', 'is_parent'])


def get_port(port_name):
    """Retrieves the devlink port information, including the PF name."""
    ports = priv_devlink.get_port_list()
    # Build port index, with PF reference and VF index.
    port_indexes = []
    ret = None
    for port in ports:
        pf_pci = linux_utils.get_attr(port, 'DEVLINK_ATTR_DEV_NAME')
        name = linux_utils.get_attr(port, 'DEVLINK_ATTR_PORT_NETDEV_NAME')
        index = linux_utils.get_attr(port, 'DEVLINK_ATTR_PORT_INDEX')
        pf_num = index >> 16
        is_parent = index & 0xFFFF == 0xFFFF
        vf_num = linux_utils.get_attr(port, 'DEVLINK_ATTR_PORT_PCI_VF_NUMBER')
        port_indexes.append(PortIndex(name, pf_pci, pf_num, vf_num, is_parent))

        if name == port_name:
            ret = {'pf_pci': pf_pci,
                   'pf_num': pf_num,
                   'pf_name': None,
                   'vf_num': vf_num,
                   'vf_name': name,
                   }

    if ret:
        for port in port_indexes:
            if port.pf_num == ret['pf_num'] and port.is_parent:
                ret['pf_name'] = port.name

    return ret



# Copyright 2023 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op
import sqlalchemy as sa


"""remove in_use from subnet

Revision ID: 93f394357a27
Revises: fc153938cdc1
Create Date: 2023-03-07 14:48:15.763633

"""

# revision identifiers, used by Alembic.
revision = '93f394357a27'
down_revision = 'fc153938cdc1'


def upgrade():
    op.drop_column('subnets', 'in_use')


def expand_drop_exceptions():
    """Support dropping 'in_use' column in table 'subnets'"""

    return {
        sa.Column: ['subnets.in_use']
    }



# Copyright (c) 2013 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random

from neutron_lib.agent import topics
from neutron_lib.api import extensions
from neutron_lib import constants
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from neutron_lib import rpc as n_rpc
from oslo_log import log as logging
import oslo_messaging

from neutron.api.rpc.agentnotifiers import utils as ag_utils
from neutron.common import utils


LOG = logging.getLogger(__name__)

# default messaging timeout is 60 sec, so 2 here is chosen to not block API
# call for more than 2 minutes
AGENT_NOTIFY_MAX_ATTEMPTS = 2


class L3AgentNotifyAPI(object):
    """API for plugin to notify L3 agent."""

    def __init__(self, topic=topics.L3_AGENT):
        target = oslo_messaging.Target(topic=topic, version='1.0')
        self.client = n_rpc.get_client(target)

    @utils.disable_notifications
    def _notification_host(self, context, method, host, use_call=False,
                           **kwargs):
        """Notify the agent that is hosting the router."""
        LOG.debug('Notify agent at %(host)s the message '
                  '%(method)s', {'host': host,
                                 'method': method})
        cctxt = self.client.prepare(server=host)
        rpc_method = (ag_utils.retry(cctxt.call, AGENT_NOTIFY_MAX_ATTEMPTS)
                      if use_call else cctxt.cast)
        rpc_method(context, method, **kwargs)

    @utils.disable_notifications
    def _agent_notification(self, context, method, router_ids, operation,
                            shuffle_agents):
        """Notify changed routers to hosting l3 agents."""
        adminContext = context if context.is_admin else context.elevated()
        plugin = directory.get_plugin(plugin_constants.L3)
        for router_id in router_ids:
            hosts = plugin.get_hosts_to_notify(adminContext, router_id)
            if shuffle_agents:
                random.shuffle(hosts)
            for host in hosts:
                LOG.debug('Notify agent at %(topic)s.%(host)s the message '
                          '%(method)s',
                          {'topic': topics.L3_AGENT,
                           'host': host,
                           'method': method})
                cctxt = self.client.prepare(topic=topics.L3_AGENT,
                                            server=host,
                                            version='1.1')
                cctxt.cast(context, method, routers=[router_id])

    @utils.disable_notifications
    def _agent_notification_arp(self, context, method, router_id,
                                operation, data):
        """Notify arp details to l3 agents hosting router."""
        if not router_id:
            return
        dvr_arptable = {'router_id': router_id, 'arp_table': data}
        LOG.debug('Fanout dvr_arptable update: %s', dvr_arptable)
        cctxt = self.client.prepare(fanout=True, version='1.2')
        cctxt.cast(context, method, payload=dvr_arptable)

    @utils.disable_notifications
    def _notification(self, context, method, router_ids, operation,
                      shuffle_agents, schedule_routers=True):
        """Notify all the agents that are hosting the routers."""
        plugin = directory.get_plugin(plugin_constants.L3)
        if not plugin:
            LOG.error('No plugin for L3 routing registered. Cannot notify '
                      'agents with the message %s', method)
            return
        if extensions.is_extension_supported(
                plugin, constants.L3_AGENT_SCHEDULER_EXT_ALIAS):
            adminContext = (context.is_admin and
                            context or context.elevated())
            if schedule_routers:
                plugin.schedule_routers(adminContext, router_ids)
            self._agent_notification(
                context, method, router_ids, operation, shuffle_agents)
        else:
            cctxt = self.client.prepare(fanout=True)
            cctxt.cast(context, method, routers=router_ids)

    @utils.disable_notifications
    def _notification_fanout(self, context, method, router_id=None, **kwargs):
        """Fanout the information to all L3 agents.

        This function will fanout the router_id or ext_net_id
        to the L3 Agents.
        """
        ext_net_id = kwargs.get('ext_net_id')
        if router_id:
            kwargs['router_id'] = router_id
            LOG.debug('Fanout notify agent at %(topic)s the message '
                      '%(method)s on router %(router_id)s',
                      {'topic': topics.L3_AGENT,
                       'method': method,
                       'router_id': router_id})
        if ext_net_id:
            LOG.debug('Fanout notify agent at %(topic)s the message '
                      '%(method)s for external_network  %(ext_net_id)s',
                      {'topic': topics.L3_AGENT,
                       'method': method,
                       'ext_net_id': ext_net_id})
        cctxt = self.client.prepare(fanout=True)
        cctxt.cast(context, method, **kwargs)

    def agent_updated(self, context, admin_state_up, host):
        self._notification_host(context, 'agent_updated', host,
                                payload={'admin_state_up': admin_state_up})

    def router_deleted(self, context, router_id):
        self._notification_fanout(context, 'router_deleted', router_id)

    def routers_updated(self, context, router_ids, operation=None, data=None,
                        shuffle_agents=False, schedule_routers=True):
        if router_ids:
            self._notification(context, 'routers_updated', router_ids,
                               operation, shuffle_agents, schedule_routers)

    def add_arp_entry(self, context, router_id, arp_table, operation=None):
        self._agent_notification_arp(context, 'add_arp_entry', router_id,
                                     operation, arp_table)

    def del_arp_entry(self, context, router_id, arp_table, operation=None):
        self._agent_notification_arp(context, 'del_arp_entry', router_id,
                                     operation, arp_table)

    def delete_fipnamespace_for_ext_net(self, context, ext_net_id):
        self._notification_fanout(
            context, 'fipnamespace_delete_on_ext_net',
            ext_net_id=ext_net_id)

    def router_removed_from_agent(self, context, router_id, host):
        self._notification_host(context, 'router_removed_from_agent', host,
                                payload={'router_id': router_id})

    def router_added_to_agent(self, context, router_ids, host):
        # need to use call here as we want to be sure agent received
        # notification and router will not be "lost". However using call()
        # itself is not a guarantee, calling code should handle exceptions and
        # retry
        self._notification_host(context, 'router_added_to_agent', host,
                                use_call=True, payload=router_ids)

    def routers_updated_on_host(self, context, router_ids, host):
        self._notification_host(context, 'routers_updated', host,
                                routers=router_ids)



# Copyright (c) 2021 Ericsson Software Technology
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import converters
from neutron_lib import context
from neutron_lib.db import api as db_api
from oslo_config import cfg
from oslo_db import options as db_options
from oslo_log import log as logging

from neutron.common import config as common_config
from neutron.objects import ports as port_obj
from neutron.objects.qos import binding as qos_binding_obj
from neutron.objects.qos import rule as qos_rule_obj


LOG = logging.getLogger(__name__)


def setup_conf():
    conf = cfg.CONF
    common_config.register_common_config_options()
    db_group, neutron_db_opts = db_options.list_opts()[0]
    cfg.CONF.register_cli_opts(neutron_db_opts, db_group)
    conf()


def main():
    """Main method for sanitizing "ml2_port_bindings.profile" column.

    This script will sanitize "ml2_port_bindings.profile" columns existing in
    the database. In Yoga release the format of this column has changed from:
        {'allocation': '<rp_uuid>'}
    to:
        {'allocation': {'<group_uuid>': '<rp_uuid>'}}

    where group_uuid is generated based on port_id and ID of QoS rules
    belonging to that group.
    """
    setup_conf()
    admin_ctx = context.get_admin_context()
    with db_api.CONTEXT_WRITER.using(admin_ctx):
        for port_binding in port_obj.PortBinding.get_objects(admin_ctx):
            # NOTE(przszc): Before minimum packet rate rule was introduced,
            # binding-profile.allocation attribute could contain only a single
            # RP UUID, responsible for providing minimum bandwidth resources.
            # Because of that, whenever we find allocation attribute that still
            # uses old format, we can safely assume that we need to generate
            # minimum bandwidth group UUID.
            allocation = port_binding.profile.get('allocation')
            if (not allocation or isinstance(allocation, dict)):
                continue

            qos_port_binding = qos_binding_obj.QosPolicyPortBinding.get_object(
                admin_ctx, port_id=port_binding.port_id)
            if not qos_port_binding:
                LOG.error(
                    'Failed to sanitize binding-profile.allocation attribute '
                    '%s for port %s: Did not find associated QoS policy.',
                    allocation, port_binding.port_id)
                continue

            min_bw_rules = qos_rule_obj.QosMinimumBandwidthRule.get_objects(
                admin_ctx, qos_policy_id=qos_port_binding.policy_id)
            if not min_bw_rules:
                LOG.error(
                    'Failed to sanitize binding-profile.allocation attribute '
                    '%s for port %s: Associated QoS policy %s has no minimum '
                    'bandwidth rules.', allocation, port_binding.port_id,
                    qos_port_binding.policy_id)
                continue

            port_binding.profile = {
                'allocation':
                converters.convert_to_sanitized_binding_profile_allocation(
                    allocation, port_binding.port_id, min_bw_rules)}
            LOG.info('Port %s updated, New binding-profile.allocation format: '
                     '%s', port_binding.port_id, port_binding.profile)
            port_binding.update()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import dns as dns_apidef
from neutron_lib.api.definitions import dns_domain_ports as ports_apidef
from neutron_lib.api.definitions import subnet_dns_publish_fixed_ip as sn_dns
from neutron_lib.api import validators
from oslo_log import log as logging

from neutron.objects import subnet as subnet_obj
from neutron.plugins.ml2.extensions import dns_integration as dns_int

LOG = logging.getLogger(__name__)


class SubnetDNSPublishFixedIPExtensionDriver(
        dns_int.DNSDomainPortsExtensionDriver):

    _supported_extension_aliases = [dns_apidef.ALIAS,
                                    ports_apidef.ALIAS,
                                    sn_dns.ALIAS]

    def initialize(self):
        LOG.info("SubnetDNSPublishFixedIPExtensionDriver initialization "
                 "complete")

    @property
    def extension_aliases(self):
        return self._supported_extension_aliases

    def extend_subnet_dict(self, session, db_data, response_data):
        # TODO(jh): This returns None instead of the proper response_data
        # response_data = (
        #     super(SubnetDNSPublishFixedIPExtensionDriver,
        #           self).extend_subnet_dict(
        #               session, db_data, response_data))
        response_data['dns_publish_fixed_ip'] = False
        if db_data.dns_publish_fixed_ip:
            response_data['dns_publish_fixed_ip'] = True
        return response_data

    def process_create_subnet(self, plugin_context, request_data, db_data):
        flag = request_data.get(sn_dns.DNS_PUBLISH_FIXED_IP)
        if not validators.is_attr_set(flag):
            return

        if flag:
            subnet_obj.SubnetDNSPublishFixedIP(
                plugin_context,
                subnet_id=db_data['id'],
                dns_publish_fixed_ip=flag).create()
        db_data[sn_dns.DNS_PUBLISH_FIXED_IP] = flag

    def process_update_subnet(self, plugin_context, request_data, db_data):
        new_value = request_data.get(sn_dns.DNS_PUBLISH_FIXED_IP)
        if not validators.is_attr_set(new_value):
            return

        current_value = db_data.get(sn_dns.DNS_PUBLISH_FIXED_IP)
        if current_value == new_value:
            return

        subnet_id = db_data['id']
        if new_value:
            subnet_obj.SubnetDNSPublishFixedIP(
                plugin_context,
                subnet_id=subnet_id,
                dns_publish_fixed_ip=new_value).create()
        else:
            sn_obj = subnet_obj.SubnetDNSPublishFixedIP.get_object(
                plugin_context,
                subnet_id=subnet_id)
            sn_obj.delete()
        db_data[sn_dns.DNS_PUBLISH_FIXED_IP] = new_value



# Copyright 2018 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import errno
import socket

from neutron_lib import constants as n_constants
import pyroute2
from pyroute2 import protocols as pyroute2_protocols

from neutron._i18n import _
from neutron import privileged
from neutron.privileged.agent import linux as priv_linux
from neutron.privileged.agent.linux import ip_lib


_IP_VERSION_FAMILY_MAP = {n_constants.IP_VERSION_4: socket.AF_INET,
                          n_constants.IP_VERSION_6: socket.AF_INET6}


class TrafficControlClassNotFound(RuntimeError):
    message = _('Traffic control class %(classid)s not found in namespace '
                '%(namespace)s.')

    def __init__(self, message=None, classid=None, namespace=None):
        message = message or self.message % {
            'classid': classid, 'namespace': namespace}
        super(TrafficControlClassNotFound, self).__init__(message)


@privileged.default.entrypoint
def add_tc_qdisc(device, namespace=None, **kwargs):
    """Add TC qdisc"""
    index = ip_lib.get_link_id(device, namespace)
    try:
        with ip_lib.get_iproute(namespace) as ip:
            ip.tc('replace', index=index, **kwargs)
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise ip_lib.NetworkNamespaceNotFound(netns_name=namespace)
        raise


@privileged.default.entrypoint
def list_tc_qdiscs(device, namespace=None):
    """List all TC qdiscs of a device"""
    index = ip_lib.get_link_id(device, namespace)
    try:
        with ip_lib.get_iproute(namespace) as ip:
            return priv_linux.make_serializable(ip.get_qdiscs(index=index))
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise ip_lib.NetworkNamespaceNotFound(netns_name=namespace)
        raise


@privileged.default.entrypoint
def delete_tc_qdisc(device, parent=None, kind=None, namespace=None,
                    raise_interface_not_found=True,
                    raise_qdisc_not_found=True):
    """Delete a TC qdisc of a device"""
    try:
        index = ip_lib.get_link_id(device, namespace)
        args = {}
        if parent:
            args['parent'] = parent
        if kind:
            args['kind'] = kind
        with ip_lib.get_iproute(namespace) as ip:
            ip.tc('del', index=index, **args)
    except ip_lib.NetworkInterfaceNotFound:
        if raise_interface_not_found:
            raise
    except pyroute2.NetlinkError as e:
        # NOTE(ralonsoh): tc delete will raise a NetlinkError exception with
        # code (22, 'Invalid argument') if kind='ingress' and the qdisc does
        # not exist. This behaviour must be refactored in pyroute2.
        if ((e.code == errno.ENOENT or
             (e.code == errno.EINVAL and kind == 'ingress')) and
                raise_qdisc_not_found is False):
            # NOTE(ralonsoh): return error code for testing purposes
            return e.code
        raise
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise ip_lib.NetworkNamespaceNotFound(netns_name=namespace)
        raise


@privileged.default.entrypoint
def add_tc_policy_class(device, parent, classid, class_type, namespace=None,
                        **kwargs):
    """Add/replace TC policy class"""
    try:
        index = ip_lib.get_link_id(device, namespace)
        with ip_lib.get_iproute(namespace) as ip:
            ip.tc('replace-class', kind=class_type, index=index,
                  handle=classid, parent=parent, **kwargs)
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise ip_lib.NetworkNamespaceNotFound(netns_name=namespace)
        raise


@privileged.default.entrypoint
def list_tc_policy_classes(device, namespace=None):
    """List all TC policy classes of a device"""
    try:
        index = ip_lib.get_link_id(device, namespace)
        with ip_lib.get_iproute(namespace) as ip:
            return priv_linux.make_serializable(ip.get_classes(index=index))
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise ip_lib.NetworkNamespaceNotFound(netns_name=namespace)
        raise


@privileged.default.entrypoint
def delete_tc_policy_class(device, parent, classid, namespace=None,
                           **kwargs):
    """Delete TC policy class"""
    try:
        index = ip_lib.get_link_id(device, namespace)
        with ip_lib.get_iproute(namespace) as ip:
            ip.tc('del-class', index=index, handle=classid, parent=parent,
                  **kwargs)
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise ip_lib.NetworkNamespaceNotFound(netns_name=namespace)
        raise
    except pyroute2.NetlinkError as e:
        if e.code == errno.ENOENT:
            raise TrafficControlClassNotFound(classid=classid,
                                              namespace=namespace)


@privileged.default.entrypoint
def add_tc_filter_match32(device, parent, priority, class_id, keys,
                          protocol=None, namespace=None, **kwargs):
    """Add TC filter, type: match u32"""
    # NOTE(ralonsoh): by default (protocol=None), every packet is filtered.
    protocol = protocol or pyroute2_protocols.ETH_P_ALL
    try:
        index = ip_lib.get_link_id(device, namespace)
        with ip_lib.get_iproute(namespace) as ip:
            ip.tc('add-filter', kind='u32', index=index,
                  parent=parent, prio=priority, target=class_id,
                  protocol=protocol, keys=keys, **kwargs)
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise ip_lib.NetworkNamespaceNotFound(netns_name=namespace)
        raise


@privileged.default.entrypoint
def add_tc_filter_policy(device, parent, priority, rate, burst, mtu, action,
                         protocol=None, keys=None, flowid=1, namespace=None,
                         **kwargs):
    """Add TC filter, type: policy filter

    By default (protocol=None), that means every packet is shaped. "keys"
    and "target" (flowid) parameters are mandatory. If the filter is
    applied on a classless qdisc, "target" is irrelevant and a default value
    can be passed. If all packets must be shaped, an empty filter ("keys")
    can be passed.
    """
    keys = keys if keys else ['0x0/0x0']
    protocol = protocol or pyroute2_protocols.ETH_P_ALL
    try:
        index = ip_lib.get_link_id(device, namespace)
        with ip_lib.get_iproute(namespace) as ip:
            ip.tc('add-filter', kind='u32', index=index,
                  parent=parent, prio=priority, protocol=protocol,
                  rate=rate, burst=burst, mtu=mtu, action=action,
                  keys=keys, target=flowid, **kwargs)
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise ip_lib.NetworkNamespaceNotFound(netns_name=namespace)
        raise


@privileged.default.entrypoint
def list_tc_filters(device, parent, namespace=None, **kwargs):
    """List TC filters"""
    try:
        index = ip_lib.get_link_id(device, namespace)
        with ip_lib.get_iproute(namespace) as ip:
            return priv_linux.make_serializable(
                ip.get_filters(index=index, parent=parent, **kwargs))
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise ip_lib.NetworkNamespaceNotFound(netns_name=namespace)
        raise



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron.tests import base

from neutron.services.qos.drivers.ovn import driver


context = 'context'


class TestOVNQosDriver(base.BaseTestCase):

    def setUp(self):
        super(TestOVNQosDriver, self).setUp()
        self.mech_driver = mock.Mock()
        self.mech_driver._ovn_client = mock.Mock()
        self.mech_driver._ovn_client._qos_driver = mock.Mock()
        self.driver = driver.OVNQosDriver.create(self.mech_driver)
        self.policy = "policy"

    def test_create_policy(self):
        self.driver.create_policy(context, self.policy)
        self.driver._driver._ovn_client._qos_driver.create_policy.\
            assert_not_called()

    def test_update_policy(self):
        self.driver.update_policy(context, self.policy)
        self.driver._driver._ovn_client._qos_driver.update_policy.\
            assert_called_once_with(context, self.policy)

    def test_delete_policy(self):
        self.driver.delete_policy(context, self.policy)
        self.driver._driver._ovn_client._qos_driver.delete_policy.\
            assert_not_called()



# Copyright 2012 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants as n_const
from neutron_lib.plugins.ml2 import ovs_constants
from oslo_config import cfg

from neutron._i18n import _
from neutron.conf.agent import common
from neutron.conf.agent.metadata import config as meta_conf


DEFAULT_BRIDGE_MAPPINGS = []
DEFAULT_TUNNEL_TYPES = []

ovs_opts = [
    cfg.StrOpt('integration_bridge', default='br-int',
               help=_("Integration bridge to use. "
                      "Do not change this parameter unless you have a good "
                      "reason to. This is the name of the OVS integration "
                      "bridge. There is one per hypervisor. The integration "
                      "bridge acts as a virtual 'patch bay'. All VM VIFs are "
                      "attached to this bridge and then 'patched' according "
                      "to their network connectivity.")),
    cfg.StrOpt('tunnel_bridge', default='br-tun',
               help=_("Tunnel bridge to use.")),
    cfg.StrOpt('int_peer_patch_port', default='patch-tun',
               help=_("Peer patch port in integration bridge for tunnel "
                      "bridge.")),
    cfg.StrOpt('tun_peer_patch_port', default='patch-int',
               help=_("Peer patch port in tunnel bridge for integration "
                      "bridge.")),
    cfg.IPOpt('local_ip',
              help=_("IP address of local overlay (tunnel) network endpoint. "
                     "Use either an IPv4 or IPv6 address that resides on one "
                     "of the host network interfaces. The IP version of this "
                     "value must match the value of the 'overlay_ip_version' "
                     "option in the ML2 plug-in configuration file on the "
                     "neutron server node(s).")),
    cfg.ListOpt('bridge_mappings',
                default=DEFAULT_BRIDGE_MAPPINGS,
                help=_("Comma-separated list of <physical_network>:<bridge> "
                       "tuples mapping physical network names to the agent's "
                       "node-specific Open vSwitch bridge names to be used "
                       "for flat and VLAN networks. The length of bridge "
                       "names should be no more than 11. Each bridge must "
                       "exist, and should have a physical network interface "
                       "configured as a port. All physical networks "
                       "configured on the server should have mappings to "
                       "appropriate bridges on each agent. "
                       "Note: If you remove a bridge from this "
                       "mapping, make sure to disconnect it from the "
                       "integration bridge as it won't be managed by the "
                       "agent anymore.")),
    cfg.ListOpt('resource_provider_bandwidths',
                default=[],
                help=_("Comma-separated list of "
                       "<bridge>:<egress_bw>:<ingress_bw> tuples, showing "
                       "the available bandwidth for the given bridge in the "
                       "given direction. The direction is meant from VM "
                       "perspective. Bandwidth is measured in kilobits per "
                       "second (kbps). The bridge must appear in "
                       "bridge_mappings as the value. But not all bridges in "
                       "bridge_mappings must be listed here. For a bridge not "
                       "listed here we neither create a resource provider in "
                       "placement nor report inventories against. An omitted "
                       "direction means we do not report an inventory for the "
                       "corresponding class.")),
    cfg.DictOpt('resource_provider_hypervisors',
                default={},
                help=_("Mapping of bridges to hypervisors: "
                       "<bridge>:<hypervisor>,... "
                       "hypervisor name is used to locate the parent of the "
                       "resource provider tree. Only needs to be set in the "
                       "rare case when the hypervisor name is different from "
                       "the resource_provider_default_hypervisor config "
                       "option value as known by the nova-compute managing "
                       "that hypervisor.")),
    cfg.ListOpt('resource_provider_packet_processing_without_direction',
                default=[],
                help=_("Comma-separated list of "
                       "<hypervisor>:<packet_rate> tuples, defining the "
                       "minimum packet rate the OVS backend can guarantee in "
                       "kilo (1000) packet per second. The hypervisor name is "
                       "used to locate the parent of the resource provider "
                       "tree. Only needs to be set in the rare case when the "
                       "hypervisor name is different from the DEFAULT.host "
                       "config option value as known by the nova-compute "
                       "managing that hypervisor or if multiple hypervisors "
                       "are served by the same OVS backend. The default is :0 "
                       "which means no packet processing capacity is "
                       "guaranteed on the hypervisor named according to "
                       "DEFAULT.host.")),
    cfg.ListOpt('resource_provider_packet_processing_with_direction',
                default=[],
                help=_("Similar to the "
                       "resource_provider_packet_processing_without_direction "
                       "but used in case the OVS backend has hardware offload "
                       "capabilities. In this case the format is "
                       "<hypervisor>:<egress_pkt_rate>:<ingress_pkt_rate> "
                       "which allows defining packet processing capacity per "
                       "traffic direction. The direction is meant from the VM "
                       "perspective. Note that the "
                       "resource_provider_packet_processing_without_direction "
                       "and the "
                       "resource_provider_packet_processing_with_direction "
                       "are mutually exclusive options.")),
    cfg.StrOpt('resource_provider_default_hypervisor',
               help=_("The default hypervisor name used to locate the parent "
                      "of the resource provider. If this option is not set, "
                      "canonical name is used")),
    cfg.DictOpt('resource_provider_inventory_defaults',
                default={'allocation_ratio': 1.0,
                         'min_unit': 1,
                         'step_size': 1,
                         'reserved': 0},
                help=_("Key:value pairs to specify defaults used "
                       "while reporting resource provider inventories. "
                       "Possible keys with their types: "
                       "allocation_ratio:float, "
                       "max_unit:int, min_unit:int, "
                       "reserved:int, step_size:int, "
                       "See also: "
                       "https://docs.openstack.org/api-ref/placement/"
                       "#update-resource-provider-inventories")),
    cfg.DictOpt('resource_provider_packet_processing_inventory_defaults',
                default={'allocation_ratio': 1.0,
                         'min_unit': 1,
                         'step_size': 1,
                         'reserved': 0},
                help=_("Key:value pairs to specify defaults used "
                       "while reporting packet rate inventories. "
                       "Possible keys with their types: "
                       "allocation_ratio:float, "
                       "max_unit:int, min_unit:int, "
                       "reserved:int, step_size:int, "
                       "See also: "
                       "https://docs.openstack.org/api-ref/placement/"
                       "#update-resource-provider-inventories")),
    cfg.StrOpt('datapath_type', default=ovs_constants.OVS_DATAPATH_SYSTEM,
               choices=[ovs_constants.OVS_DATAPATH_SYSTEM,
                        ovs_constants.OVS_DATAPATH_NETDEV],
               help=_("OVS datapath to use. 'system' is the default value and "
                      "corresponds to the kernel datapath. To enable the "
                      "userspace datapath set this value to 'netdev'.")),
    cfg.StrOpt('vhostuser_socket_dir',
               default=ovs_constants.VHOST_USER_SOCKET_DIR,
               help=_("OVS vhost-user socket directory.")),
    cfg.IPOpt('of_listen_address', default='127.0.0.1',
              help=_("Address to listen on for OpenFlow connections.")),
    cfg.PortOpt('of_listen_port', default=6633,
                help=_("Port to listen on for OpenFlow connections.")),
    cfg.IntOpt('of_connect_timeout', default=300,
               help=_("Timeout in seconds to wait for "
                      "the local switch connecting the controller.")),
    cfg.IntOpt('of_request_timeout', default=300,
               help=_("Timeout in seconds to wait for a single "
                      "OpenFlow request.")),
    cfg.IntOpt('of_inactivity_probe', default=10,
               help=_("The inactivity_probe interval in seconds for the local "
                      "switch connection to the controller. "
                      "A value of 0 disables inactivity probes.")),
    cfg.BoolOpt('openflow_processed_per_port',
                default=False,
                help=_('If enabled, all OpenFlow rules associated to a port '
                       'are processed at once, in one single transaction. '
                       'That avoids possible inconsistencies during OVS agent '
                       'restart and port updates. '
                       'If disabled, the flows will be processed in batches '
                       'of ``_constants.AGENT_RES_PROCESSING_STEP`` number of '
                       'OpenFlow rules.')),
]

agent_opts = [
    cfg.BoolOpt('minimize_polling',
                default=True,
                help=_("Minimize polling by monitoring OVSDB for interface "
                       "changes.")),
    cfg.IntOpt('ovsdb_monitor_respawn_interval',
               default=ovs_constants.DEFAULT_OVSDBMON_RESPAWN,
               help=_("The number of seconds to wait before respawning the "
                      "OVSDB monitor after losing communication with it.")),
    cfg.ListOpt('tunnel_types', default=DEFAULT_TUNNEL_TYPES,
                help=_("Network types supported by the agent "
                       "(gre, vxlan and/or geneve).")),
    cfg.PortOpt('vxlan_udp_port', default=n_const.VXLAN_UDP_PORT,
                help=_("The UDP port to use for VXLAN tunnels.")),
    cfg.BoolOpt('l2_population', default=False,
                help=_("Use ML2 l2population mechanism driver to learn "
                       "remote MAC and IPs and improve tunnel scalability.")),
    cfg.BoolOpt('arp_responder', default=False,
                help=_("Enable local ARP responder if it is supported. "
                       "Requires OVS 2.1 and ML2 l2population driver. "
                       "Allows the switch (when supporting an overlay) "
                       "to respond to an ARP request locally without "
                       "performing a costly ARP broadcast into the overlay. "
                       "NOTE: If enable_distributed_routing is set to True "
                       "then arp_responder will automatically be set to True "
                       "in the agent, regardless of the setting in the config "
                       "file.")),
    cfg.BoolOpt('dont_fragment', default=True,
                help=_("Set or un-set the do not fragment (DF) bit on "
                       "outgoing IP packet carrying GRE/VXLAN tunnel.")),
    cfg.BoolOpt('enable_distributed_routing', default=False,
                help=_("Make the l2 agent run in DVR mode.")),
    cfg.BoolOpt('drop_flows_on_start', default=False,
                help=_("Reset flow table on start. Setting this to True will "
                       "cause brief traffic interruption.")),
    cfg.BoolOpt('tunnel_csum', default=False,
                help=_("Set or un-set the tunnel header checksum on "
                       "outgoing IP packet carrying GRE/VXLAN tunnel.")),
    cfg.BoolOpt('baremetal_smartnic', default=False,
                help=_("Enable the agent to process Smart NIC ports.")),
    cfg.BoolOpt('explicitly_egress_direct', default=False,
                help=_("When set to True, the accepted egress unicast "
                       "traffic will not use action NORMAL. The accepted "
                       "egress packets will be taken care of in the final "
                       "egress tables direct output flows for unicast "
                       "traffic.")),
]

dhcp_opts = [
    cfg.BoolOpt('enable_ipv6', default=True,
                help=_("When set to True, the OVS agent DHCP "
                       "extension will add related flows for "
                       "DHCPv6 packets.")),
]

local_ip_opts = [
    cfg.BoolOpt('static_nat', default=False,
                help=_("When set to True, the Local IP Openvswitch agent "
                       "extension will use static NAT rules instead of using "
                       "conntrack. This allows to use feature in OVS offload "
                       "and DPDK scenarios at the cost of number and "
                       "complexity of flows. This also allows using this "
                       "feature together with the OVS firewall.")),
]


def register_ovs_agent_opts(cfg=cfg.CONF):
    cfg.register_opts(ovs_opts, "OVS")
    cfg.register_opts(agent_opts, "AGENT")
    cfg.register_opts(dhcp_opts, "DHCP")
    cfg.register_opts(common.DHCP_PROTOCOL_OPTS, "DHCP")
    cfg.register_opts(local_ip_opts, "LOCAL_IP")
    cfg.register_opts(meta_conf.METADATA_PROXY_HANDLER_OPTS, "METADATA")


def register_ovs_opts(cfg=cfg.CONF):
    cfg.register_opts(ovs_opts, "OVS")



# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import sys

from neutron_lib import constants as p_const
from neutron_lib import context
from neutron_lib.db import api as db_api
from neutron_lib import exceptions as exc
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from neutron_lib.plugins.ml2 import api
from neutron_lib.plugins import utils as plugin_utils
from oslo_config import cfg
from oslo_log import log
from oslo_utils import uuidutils

from neutron._i18n import _
from neutron.conf.plugins.ml2.drivers import driver_type
from neutron.db.models.plugins.ml2 import vlanallocation as vlan_alloc_model
from neutron.objects import network_segment_range as range_obj
from neutron.objects.plugins.ml2 import vlanallocation as vlanalloc
from neutron.plugins.ml2.drivers import helpers
from neutron.services.network_segment_range import plugin as range_plugin

LOG = log.getLogger(__name__)

driver_type.register_ml2_drivers_vlan_opts()


class VlanTypeDriver(helpers.SegmentTypeDriver):
    """Manage state for VLAN networks with ML2.

    The VlanTypeDriver implements the 'vlan' network_type. VLAN
    network segments provide connectivity between VMs and other
    devices using any connected IEEE 802.1Q conformant
    physical_network segmented into virtual networks via IEEE 802.1Q
    headers. Up to 4094 VLAN network segments can exist on each
    available physical_network.
    """

    def __init__(self):
        super(VlanTypeDriver, self).__init__(vlanalloc.VlanAllocation)
        self.model_segmentation_id = vlan_alloc_model.VlanAllocation.vlan_id
        self._parse_network_vlan_ranges()

    @db_api.retry_db_errors
    def _populate_new_default_network_segment_ranges(self):
        ctx = context.get_admin_context()
        for (physical_network, vlan_ranges) in (
                self.network_vlan_ranges.items()):
            for vlan_min, vlan_max in vlan_ranges:
                res = {
                    'id': uuidutils.generate_uuid(),
                    'name': '',
                    'default': True,
                    'shared': True,
                    'network_type': p_const.TYPE_VLAN,
                    'physical_network': physical_network,
                    'minimum': vlan_min,
                    'maximum': vlan_max}
                with db_api.CONTEXT_WRITER.using(ctx):
                    new_default_range_obj = (
                        range_obj.NetworkSegmentRange(ctx, **res))
                    new_default_range_obj.create()

    def _parse_network_vlan_ranges(self):
        try:
            self.network_vlan_ranges = plugin_utils.parse_network_vlan_ranges(
                cfg.CONF.ml2_type_vlan.network_vlan_ranges)
        except Exception:
            LOG.exception("Failed to parse network_vlan_ranges. "
                          "Service terminated!")
            sys.exit(1)
        LOG.info("Network VLAN ranges: %s", self.network_vlan_ranges)

    @db_api.retry_db_errors
    def _sync_vlan_allocations(self):
        ctx = context.get_admin_context()
        with db_api.CONTEXT_WRITER.using(ctx):
            # VLAN ranges per physical network:
            #   {phy1: [(1, 10), (30, 50)], ...}
            ranges = self.get_network_segment_ranges()

            # Delete those VLAN registers from unconfigured physical networks
            physnets = vlanalloc.VlanAllocation.get_physical_networks(ctx)
            physnets_unconfigured = physnets - set(ranges)
            if physnets_unconfigured:
                LOG.debug('Removing any VLAN register on physical networks %s',
                          physnets_unconfigured)
                vlanalloc.VlanAllocation.delete_physical_networks(
                    ctx, physnets_unconfigured)

            # Get existing allocations for all configured physical networks
            allocations = collections.defaultdict(list)
            for alloc in vlanalloc.VlanAllocation.get_objects(ctx):
                allocations[alloc.physical_network].append(alloc)

            for physical_network, vlan_ranges in ranges.items():
                # determine current configured allocatable vlans for
                # this physical network
                vlan_ids = set()
                for vlan_min, vlan_max in vlan_ranges:
                    vlan_ids |= set(range(vlan_min, vlan_max + 1))

                # remove from table unallocated vlans not currently
                # allocatable
                if physical_network in allocations:
                    for alloc in allocations[physical_network]:
                        try:
                            # see if vlan is allocatable
                            vlan_ids.remove(alloc.vlan_id)
                        except KeyError:
                            # it's not allocatable, so check if its allocated
                            if not alloc.allocated:
                                # it's not, so remove it from table
                                LOG.debug("Removing vlan %(vlan_id)s on "
                                          "physical network "
                                          "%(physical_network)s from pool",
                                          {'vlan_id': alloc.vlan_id,
                                           'physical_network':
                                           physical_network})
                                # This UPDATE WHERE statement blocks anyone
                                # from concurrently changing the allocation
                                # values to True while our transaction is
                                # open so we don't accidentally delete
                                # allocated segments. If someone has already
                                # allocated, update_objects will return 0 so we
                                # don't delete.
                                if vlanalloc.VlanAllocation.update_objects(
                                        ctx, values={'allocated': False},
                                        allocated=False, vlan_id=alloc.vlan_id,
                                        physical_network=physical_network):
                                    alloc.delete()
                    del allocations[physical_network]

                # Add missing allocatable VLAN registers for "physical_network"
                vlanalloc.VlanAllocation.bulk_create(ctx, physical_network,
                                                     vlan_ids)

    @db_api.retry_db_errors
    def _get_network_segment_ranges_from_db(self):
        ranges = {}
        ctx = context.get_admin_context()
        with db_api.CONTEXT_READER.using(ctx):
            range_objs = (range_obj.NetworkSegmentRange.get_objects(
                ctx, network_type=self.get_type()))
            for obj in range_objs:
                physical_network = obj['physical_network']
                if physical_network not in ranges:
                    ranges[physical_network] = []
                ranges[physical_network].append((obj['minimum'],
                                                 obj['maximum']))
        return ranges

    def get_type(self):
        return p_const.TYPE_VLAN

    def initialize(self):
        if not range_plugin.is_network_segment_range_enabled():
            # service plugins are initialized/loaded after the ML2 driver
            # initialization. Thus, we base on the information whether
            # ``network_segment_range`` service plugin is enabled/defined in
            # ``neutron.conf`` to decide whether to skip the first time sync
            # allocation during driver initialization, instead of using the
            # directory.get_plugin() method - the normal way used elsewhere to
            # check if a plugin is loaded.
            self._sync_vlan_allocations()
        LOG.info("VlanTypeDriver initialization complete")

    def initialize_network_segment_range_support(self):
        self._delete_expired_default_network_segment_ranges()
        self._populate_new_default_network_segment_ranges()
        # Override self.network_vlan_ranges with the network segment range
        # information from DB and then do a sync_allocations since the
        # segment range service plugin has not yet been loaded at this
        # initialization time.
        self.network_vlan_ranges = self._get_network_segment_ranges_from_db()
        self._sync_vlan_allocations()

    def update_network_segment_range_allocations(self):
        self._sync_vlan_allocations()

    def get_network_segment_ranges(self):
        """Get the driver network segment ranges.

        Queries all VLAN network segment ranges from DB if the
        ``NETWORK_SEGMENT_RANGE`` service plugin is enabled. Otherwise,
        they will be loaded from the host config file - `ml2_conf.ini`.
        """
        ranges = self.network_vlan_ranges
        if directory.get_plugin(plugin_constants.NETWORK_SEGMENT_RANGE):
            ranges = self._get_network_segment_ranges_from_db()

        return ranges

    def is_partial_segment(self, segment):
        return segment.get(api.SEGMENTATION_ID) is None

    def validate_provider_segment(self, segment):
        physical_network = segment.get(api.PHYSICAL_NETWORK)
        segmentation_id = segment.get(api.SEGMENTATION_ID)
        ranges = self.get_network_segment_ranges()
        if physical_network:
            if physical_network not in ranges:
                msg = (_("physical_network '%s' unknown "
                         "for VLAN provider network") % physical_network)
                raise exc.InvalidInput(error_message=msg)
            if segmentation_id is not None:
                if not plugin_utils.is_valid_vlan_tag(segmentation_id):
                    msg = (_("segmentation_id out of range (%(min)s through "
                             "%(max)s)") %
                           {'min': p_const.MIN_VLAN_TAG,
                            'max': p_const.MAX_VLAN_TAG})
                    raise exc.InvalidInput(error_message=msg)
        elif segmentation_id is not None:
            msg = _("segmentation_id requires physical_network for VLAN "
                    "provider network")
            raise exc.InvalidInput(error_message=msg)

        for key, value in segment.items():
            if value and key not in [api.NETWORK_TYPE,
                                     api.PHYSICAL_NETWORK,
                                     api.SEGMENTATION_ID]:
                msg = _("%s prohibited for VLAN provider network") % key
                raise exc.InvalidInput(error_message=msg)

    def reserve_provider_segment(self, context, segment, filters=None):
        filters = filters or {}
        project_id = filters.get('project_id')
        filters = {}
        physical_network = segment.get(api.PHYSICAL_NETWORK)
        if physical_network is not None:
            filters['physical_network'] = physical_network
            vlan_id = segment.get(api.SEGMENTATION_ID)
            if vlan_id is not None:
                filters['vlan_id'] = vlan_id

        if self.is_partial_segment(segment):
            if (directory.get_plugin(
                    plugin_constants.NETWORK_SEGMENT_RANGE) and project_id):
                filters['project_id'] = project_id
            alloc = self.allocate_partially_specified_segment(
                context, **filters)
            if not alloc:
                raise exc.NoNetworkAvailable()
        else:
            alloc = self.allocate_fully_specified_segment(
                context, **filters)
            if not alloc:
                raise exc.VlanIdInUse(**filters)

        return {api.NETWORK_TYPE: p_const.TYPE_VLAN,
                api.PHYSICAL_NETWORK: alloc.physical_network,
                api.SEGMENTATION_ID: alloc.vlan_id,
                api.MTU: self.get_mtu(alloc.physical_network)}

    def allocate_tenant_segment(self, context, filters=None):
        filters = filters or {}
        ranges = self.get_network_segment_ranges()
        for physnet in ranges:
            filters['physical_network'] = physnet
            alloc = self.allocate_partially_specified_segment(
                context, **filters)
            if alloc:
                break
        else:
            return
        return {api.NETWORK_TYPE: p_const.TYPE_VLAN,
                api.PHYSICAL_NETWORK: alloc.physical_network,
                api.SEGMENTATION_ID: alloc.vlan_id,
                api.MTU: self.get_mtu(alloc.physical_network)}

    def release_segment(self, context, segment):
        physical_network = segment[api.PHYSICAL_NETWORK]
        vlan_id = segment[api.SEGMENTATION_ID]

        vlan_ranges = self.get_network_segment_ranges()
        ranges = vlan_ranges.get(physical_network, [])
        inside = any(lo <= vlan_id <= hi for lo, hi in ranges)
        count = False

        with db_api.CONTEXT_WRITER.using(context):
            alloc = vlanalloc.VlanAllocation.get_object(
                context, physical_network=physical_network, vlan_id=vlan_id)
            if alloc:
                if inside and alloc.allocated:
                    count = True
                    alloc.allocated = False
                    alloc.update()
                    LOG.debug("Releasing vlan %(vlan_id)s on physical "
                              "network %(physical_network)s to pool",
                              {'vlan_id': vlan_id,
                               'physical_network': physical_network})
                else:
                    count = True
                    alloc.delete()
                    LOG.debug("Releasing vlan %(vlan_id)s on physical "
                              "network %(physical_network)s outside pool",
                              {'vlan_id': vlan_id,
                               'physical_network': physical_network})

        if not count:
            LOG.warning("No vlan_id %(vlan_id)s found on physical "
                        "network %(physical_network)s",
                        {'vlan_id': vlan_id,
                         'physical_network': physical_network})

    def get_mtu(self, physical_network):
        seg_mtu = super(VlanTypeDriver, self).get_mtu()
        mtu = []
        if seg_mtu > 0:
            mtu.append(seg_mtu)
        if physical_network in self.physnet_mtus:
            mtu.append(int(self.physnet_mtus[physical_network]))
        return min(mtu) if mtu else 0
